module Core where

import Control.Exception                  (IOException)
import Control.Monad.Catch                (catch)
import Data.Aeson                         (FromJSON, ToJSON)
import qualified Data.Text                as T
import Data.ByteString.Lazy.UTF8          (fromString)
import Database.PostgreSQL.Simple         (Query)
import Database.PostgreSQL.Simple.FromRow (FromRow, field, fromRow)
import Database.PostgreSQL.Simple.ToField (ToField)
import Database.PostgreSQL.Simple.ToRow   (ToRow, toRow)
import Effectful
import Effectful.Dispatch.Dynamic
import Effectful.Error.Static             (Error, throwError)
import Effectful.FileSystem
import GHC.Generics                       (Generic)
import Servant hiding                     ((:>), throwError)
import qualified Servant                  as S

--
-- Core data
--
data User = User { userId :: Int, userName :: String}
  deriving (Show, Generic)

instance FromRow User where
    fromRow = User <$> field <*> field
instance ToRow User where
    toRow (User uid name) = toRow (uid, name)

instance ToJSON User
instance FromJSON User

--
-- Effects
--
type AppEff = Eff '[Database, FileSystem, Logger, Error ServerError, IOE]

-- Database
data Database :: Effect where
  DatabaseRead  :: (Query, Int)    -> Database (Eff es) (Maybe User)
  DatabaseWrite :: (Query, String) -> Database (Eff es) ()

type instance DispatchOf Database = 'Dynamic

type DatabaseEffects es = (IOE :> es, Logger :> es, Error ServerError :> es)

databaseRead  :: (Database :> es, Error ServerError :> es) => (Query, Int) -> Eff es (Maybe User)
databaseRead  = send . DatabaseRead

databaseWrite :: (Database :> es, Error ServerError :> es) => (Query, String) -> Eff es ()
databaseWrite = send . DatabaseWrite

-- Logger
data Logger :: Effect where
  WriteLog :: String -> Logger (Eff es) ()

type instance DispatchOf Logger = 'Dynamic

writeLog :: (Logger :> es, Error ServerError :> es) => String -> Eff es ()
writeLog = send . WriteLog

runLoggerIO :: (IOE :> es, Error ServerError :> es) => Eff (Logger : es) a -> Eff es a
runLoggerIO = interpret $ \_ -> \case
  WriteLog msg -> log msg
  where
    log :: (IOE :> es, Error ServerError :> es) => String -> Eff es ()
    log msg = adapt $ putStrLn msg

-- Utility
adapt :: (IOE :> es, Error ServerError :> es) => IO a -> Eff es a
adapt m = liftIO m `catch` \(e::IOException) -> do
    liftIO $ putStrLn $ "ERROR Exception: " ++ show e
    throwError $ ServerError
      { errHTTPCode     = 500
      , errReasonPhrase = "Internal Server Error"
      , errBody         = "This incident will be investigated."
      , errHeaders      = []
      }

--
-- Routes
--
type Root     = Get '[PlainText] T.Text
type UserList = "user" S.:> Get '[JSON] [User]
type UserGet  = "user" S.:> Capture "userId"     Int    S.:> Get  '[JSON] User
type UserPost = "user" S.:> ReqBody '[PlainText] String S.:> PostCreated '[PlainText] NoContent

type API = Root
      :<|> UserList
      :<|> UserGet
      :<|> UserPost
